/*   (C) Copyright 2008, 2009 Stijn van Dongen
 *
 * This file is part of tingea.  You can redistribute and/or modify tingea
 * under the terms of the GNU General Public License; either version 3 of the
 * License or (at your option) any later version.  You should have received a
 * copy of the GPL along with tingea, in the file COPYING.
*/

#include "types.h"

const char* mcx_status_list[]
=  {  "OK"
   ,  "FAIL"
   ,  "DONE"
   ,  "IGNORE"
   ,  "NOMEM"
   ,  "ABORT"
   ,  "NEW"
   ,  "UNUSED"
   }  ;

